@@ -4,7 +4,7 @@
 
 import logging
 from collections import defaultdict
-from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Union
+from typing import Any, Callable, Dict, Iterable, List, Mapping, MutableMapping, Optional, Tuple, Union
 from unittest.mock import call
 
 import pytest
@@ -13,17 +13,22 @@ from airbyte_cdk.models import (
     AirbyteConnectionStatus,
     AirbyteMessage,
     AirbyteRecordMessage,
+    AirbyteStateBlob,
     AirbyteStateMessage,
+    AirbyteStateType,
     AirbyteStream,
+    AirbyteStreamState,
     ConfiguredAirbyteCatalog,
     ConfiguredAirbyteStream,
     DestinationSyncMode,
     Status,
+    StreamDescriptor,
     SyncMode,
     Type,
 )
 from airbyte_cdk.sources import AbstractSource
-from airbyte_cdk.sources.streams import Stream
+from airbyte_cdk.sources.connector_state_manager import ConnectorStateManager
+from airbyte_cdk.sources.streams import IncrementalMixin, Stream
 from airbyte_cdk.utils.traced_exception import AirbyteTracedException
 
 logger = logging.getLogger("airbyte")
@@ -34,20 +39,53 @@ class MockSource(AbstractSource):
         self,
         check_lambda: Callable[[], Tuple[bool, Optional[Any]]] = None,
         streams: List[Stream] = None,
+        per_stream: bool = True,
     ):
         self._streams = streams
         self.check_lambda = check_lambda
+        self.per_stream = per_stream
 
     def check_connection(self, logger: logging.Logger, config: Mapping[str, Any]) -> Tuple[bool, Optional[Any]]:
         if self.check_lambda:
             return self.check_lambda()
-        return (False, "Missing callable.")
+        return False, "Missing callable."
 
     def streams(self, config: Mapping[str, Any]) -> List[Stream]:
         if not self._streams:
             raise Exception("Stream is not set")
         return self._streams
 
+    @property
+    def per_stream_state_enabled(self) -> bool:
+        return self.per_stream
+
+
+class StreamNoStateMethod(Stream):
+    name = "managers"
+    primary_key = None
+
+    def read_records(self, *args, **kwargs) -> Iterable[Mapping[str, Any]]:
+        return {}
+
+
+class MockStreamOverridesStateMethod(Stream, IncrementalMixin):
+    name = "teams"
+    primary_key = None
+    cursor_field = "updated_at"
+    _cursor_value = ""
+    start_date = "1984-12-12"
+
+    def read_records(self, *args, **kwargs) -> Iterable[Mapping[str, Any]]:
+        return {}
+
+    @property
+    def state(self) -> MutableMapping[str, Any]:
+        return {self.cursor_field: self._cursor_value} if self._cursor_value else {}
+
+    @state.setter
+    def state(self, value: MutableMapping[str, Any]):
+        self._cursor_value = value.get(self.cursor_field, self.start_date)
+
 
 def test_successful_check():
     """Tests that if a source returns TRUE for the connection check the appropriate connectionStatus success message is returned"""
@@ -181,6 +219,21 @@ def _as_records(stream: str, data: List[Dict[str, Any]]) -> List[AirbyteMessage]
     return [_as_record(stream, datum) for datum in data]
 
 
+def _as_state(state_data: Dict[str, Any], stream_name: str = "", per_stream_state: Dict[str, Any] = None):
+    if per_stream_state:
+        return AirbyteMessage(
+            type=Type.STATE,
+            state=AirbyteStateMessage(
+                type=AirbyteStateType.STREAM,
+                stream=AirbyteStreamState(
+                    stream_descriptor=StreamDescriptor(name=stream_name), stream_state=AirbyteStateBlob.parse_obj(per_stream_state)
+                ),
+                data=state_data,
+            ),
+        )
+    return AirbyteMessage(type=Type.STATE, state=AirbyteStateMessage(data=state_data))
+
+
 def _configured_stream(stream: Stream, sync_mode: SyncMode):
     return ConfiguredAirbyteStream(
         stream=stream.as_airbyte_stream(),
@@ -249,17 +302,39 @@ def test_valid_full_refresh_read_with_slices(mocker):
     assert expected == messages
 
 
-def _state(state_data: Dict[str, Any]):
-    return AirbyteMessage(type=Type.STATE, state=AirbyteStateMessage(data=state_data))
-
-
 class TestIncrementalRead:
-    def test_with_state_attribute(self, mocker):
+    @pytest.mark.parametrize(
+        "use_legacy",
+        [
+            pytest.param(True, id="test_incoming_stream_state_as_legacy_format"),
+            pytest.param(False, id="test_incoming_stream_state_as_per_stream_format"),
+        ],
+    )
+    @pytest.mark.parametrize(
+        "per_stream_enabled",
+        [
+            pytest.param(True, id="test_source_emits_state_as_per_stream_format"),
+            pytest.param(False, id="test_source_emits_state_as_per_stream_format"),
+        ],
+    )
+    def test_with_state_attribute(self, mocker, use_legacy, per_stream_enabled):
         """Test correct state passing for the streams that have a state attribute"""
         stream_output = [{"k1": "v1"}, {"k2": "v2"}]
         old_state = {"cursor": "old_value"}
-        new_state = {"cursor": "new_value"}
-        s1 = MockStreamWithState(
+        if use_legacy:
+            input_state = {"s1": old_state}
+        else:
+            input_state = [
+                AirbyteStateMessage(
+                    type=AirbyteStateType.STREAM,
+                    stream=AirbyteStreamState(
+                        stream_descriptor=StreamDescriptor(name="s1"), stream_state=AirbyteStateBlob.parse_obj(old_state)
+                    ),
+                ),
+            ]
+        new_state_from_connector = {"cursor": "new_value"}
+
+        stream_1 = MockStreamWithState(
             [
                 (
                     {"sync_mode": SyncMode.incremental, "stream_state": old_state},
@@ -268,7 +343,7 @@ class TestIncrementalRead:
             ],
             name="s1",
         )
-        s2 = MockStreamWithState(
+        stream_2 = MockStreamWithState(
             [({"sync_mode": SyncMode.incremental, "stream_state": {}}, stream_output)],
             name="s2",
         )
@@ -277,44 +352,67 @@ class TestIncrementalRead:
             MockStreamWithState,
             "state",
             new_callable=mocker.PropertyMock,
-            return_value=new_state,
+            return_value=new_state_from_connector,
         )
         mocker.patch.object(MockStreamWithState, "get_json_schema", return_value={})
-        src = MockSource(streams=[s1, s2])
+        src = MockSource(streams=[stream_1, stream_2], per_stream=per_stream_enabled)
         catalog = ConfiguredAirbyteCatalog(
             streams=[
-                _configured_stream(s1, SyncMode.incremental),
-                _configured_stream(s2, SyncMode.incremental),
+                _configured_stream(stream_1, SyncMode.incremental),
+                _configured_stream(stream_2, SyncMode.incremental),
             ]
         )
 
         expected = [
             _as_record("s1", stream_output[0]),
             _as_record("s1", stream_output[1]),
-            _state({"s1": new_state}),
+            _as_state({"s1": new_state_from_connector}, "s1", new_state_from_connector)
+            if per_stream_enabled
+            else _as_state({"s1": new_state_from_connector}),
             _as_record("s2", stream_output[0]),
             _as_record("s2", stream_output[1]),
-            _state({"s1": new_state, "s2": new_state}),
+            _as_state({"s1": new_state_from_connector, "s2": new_state_from_connector}, "s2", new_state_from_connector)
+            if per_stream_enabled
+            else _as_state({"s1": new_state_from_connector, "s2": new_state_from_connector}),
         ]
-        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state={"s1": old_state})))
+        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state=input_state)))
 
-        assert expected == messages
+        assert messages == expected
         assert state_property.mock_calls == [
             call(old_state),  # set state for s1
             call(),  # get state in the end of slice for s1
             call(),  # get state in the end of slice for s2
         ]
 
-    def test_with_checkpoint_interval(self, mocker):
+    @pytest.mark.parametrize(
+        "use_legacy",
+        [
+            pytest.param(True, id="test_incoming_stream_state_as_legacy_format"),
+            pytest.param(False, id="test_incoming_stream_state_as_per_stream_format"),
+        ],
+    )
+    @pytest.mark.parametrize(
+        "per_stream_enabled",
+        [
+            pytest.param(True, id="test_source_emits_state_as_per_stream_format"),
+            pytest.param(False, id="test_source_emits_state_as_per_stream_format"),
+        ],
+    )
+    def test_with_checkpoint_interval(self, mocker, use_legacy, per_stream_enabled):
         """Tests that an incremental read which doesn't specify a checkpoint interval outputs a STATE message
         after reading N records within a stream.
         """
+        if use_legacy:
+            input_state = defaultdict(dict)
+        else:
+            input_state = []
         stream_output = [{"k1": "v1"}, {"k2": "v2"}]
-        s1 = MockStream(
+
+        stream_1 = MockStream(
             [({"sync_mode": SyncMode.incremental, "stream_state": {}}, stream_output)],
             name="s1",
         )
-        s2 = MockStream(
+        stream_2 = MockStream(
             [({"sync_mode": SyncMode.incremental, "stream_state": {}}, stream_output)],
             name="s2",
         )
@@ -330,40 +428,59 @@ class TestIncrementalRead:
             return_value=1,
         )
 
-        src = MockSource(streams=[s1, s2])
+        src = MockSource(streams=[stream_1, stream_2], per_stream=per_stream_enabled)
         catalog = ConfiguredAirbyteCatalog(
             streams=[
-                _configured_stream(s1, SyncMode.incremental),
-                _configured_stream(s2, SyncMode.incremental),
+                _configured_stream(stream_1, SyncMode.incremental),
+                _configured_stream(stream_2, SyncMode.incremental),
             ]
         )
 
         expected = [
             _as_record("s1", stream_output[0]),
-            _state({"s1": state}),
+            _as_state({"s1": state}, "s1", state) if per_stream_enabled else _as_state({"s1": state}),
             _as_record("s1", stream_output[1]),
-            _state({"s1": state}),
-            _state({"s1": state}),
+            _as_state({"s1": state}, "s1", state) if per_stream_enabled else _as_state({"s1": state}),
+            _as_state({"s1": state}, "s1", state) if per_stream_enabled else _as_state({"s1": state}),
             _as_record("s2", stream_output[0]),
-            _state({"s1": state, "s2": state}),
+            _as_state({"s1": state, "s2": state}, "s2", state) if per_stream_enabled else _as_state({"s1": state, "s2": state}),
             _as_record("s2", stream_output[1]),
-            _state({"s1": state, "s2": state}),
-            _state({"s1": state, "s2": state}),
+            _as_state({"s1": state, "s2": state}, "s2", state) if per_stream_enabled else _as_state({"s1": state, "s2": state}),
+            _as_state({"s1": state, "s2": state}, "s2", state) if per_stream_enabled else _as_state({"s1": state, "s2": state}),
         ]
-        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state=defaultdict(dict))))
+        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state=input_state)))
 
         assert expected == messages
 
-    def test_with_no_interval(self, mocker):
+    @pytest.mark.parametrize(
+        "use_legacy",
+        [
+            pytest.param(True, id="test_incoming_stream_state_as_legacy_format"),
+            pytest.param(False, id="test_incoming_stream_state_as_per_stream_format"),
+        ],
+    )
+    @pytest.mark.parametrize(
+        "per_stream_enabled",
+        [
+            pytest.param(True, id="test_source_emits_state_as_per_stream_format"),
+            pytest.param(False, id="test_source_emits_state_as_per_stream_format"),
+        ],
+    )
+    def test_with_no_interval(self, mocker, use_legacy, per_stream_enabled):
         """Tests that an incremental read which doesn't specify a checkpoint interval outputs
         a STATE message only after fully reading the stream and does not output any STATE messages during syncing the stream.
         """
+        if use_legacy:
+            input_state = defaultdict(dict)
+        else:
+            input_state = []
         stream_output = [{"k1": "v1"}, {"k2": "v2"}]
-        s1 = MockStream(
+
+        stream_1 = MockStream(
             [({"sync_mode": SyncMode.incremental, "stream_state": {}}, stream_output)],
             name="s1",
         )
-        s2 = MockStream(
+        stream_2 = MockStream(
             [({"sync_mode": SyncMode.incremental, "stream_state": {}}, stream_output)],
             name="s2",
         )
@@ -372,30 +489,49 @@ class TestIncrementalRead:
         mocker.patch.object(MockStream, "supports_incremental", return_value=True)
         mocker.patch.object(MockStream, "get_json_schema", return_value={})
 
-        src = MockSource(streams=[s1, s2])
+        src = MockSource(streams=[stream_1, stream_2], per_stream=per_stream_enabled)
         catalog = ConfiguredAirbyteCatalog(
             streams=[
-                _configured_stream(s1, SyncMode.incremental),
-                _configured_stream(s2, SyncMode.incremental),
+                _configured_stream(stream_1, SyncMode.incremental),
+                _configured_stream(stream_2, SyncMode.incremental),
             ]
         )
 
         expected = [
             *_as_records("s1", stream_output),
-            _state({"s1": state}),
+            _as_state({"s1": state}, "s1", state) if per_stream_enabled else _as_state({"s1": state}),
             *_as_records("s2", stream_output),
-            _state({"s1": state, "s2": state}),
+            _as_state({"s1": state, "s2": state}, "s2", state) if per_stream_enabled else _as_state({"s1": state, "s2": state}),
         ]
 
-        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state=defaultdict(dict))))
+        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state=input_state)))
 
         assert expected == messages
 
-    def test_with_slices(self, mocker):
+    @pytest.mark.parametrize(
+        "use_legacy",
+        [
+            pytest.param(True, id="test_incoming_stream_state_as_legacy_format"),
+            pytest.param(False, id="test_incoming_stream_state_as_per_stream_format"),
+        ],
+    )
+    @pytest.mark.parametrize(
+        "per_stream_enabled",
+        [
+            pytest.param(True, id="test_source_emits_state_as_per_stream_format"),
+            pytest.param(False, id="test_source_emits_state_as_per_stream_format"),
+        ],
+    )
+    def test_with_slices(self, mocker, use_legacy, per_stream_enabled):
         """Tests that an incremental read which uses slices outputs each record in the slice followed by a STATE message, for each slice"""
+        if use_legacy:
+            input_state = defaultdict(dict)
+        else:
+            input_state = []
         slices = [{"1": "1"}, {"2": "2"}]
         stream_output = [{"k1": "v1"}, {"k2": "v2"}, {"k3": "v3"}]
-        s1 = MockStream(
+
+        stream_1 = MockStream(
             [
                 (
                     {
@@ -409,7 +545,7 @@ class TestIncrementalRead:
             ],
             name="s1",
         )
-        s2 = MockStream(
+        stream_2 = MockStream(
             [
                 (
                     {
@@ -429,42 +565,60 @@ class TestIncrementalRead:
         mocker.patch.object(MockStream, "get_json_schema", return_value={})
         mocker.patch.object(MockStream, "stream_slices", return_value=slices)
 
-        src = MockSource(streams=[s1, s2])
+        src = MockSource(streams=[stream_1, stream_2], per_stream=per_stream_enabled)
         catalog = ConfiguredAirbyteCatalog(
             streams=[
-                _configured_stream(s1, SyncMode.incremental),
-                _configured_stream(s2, SyncMode.incremental),
+                _configured_stream(stream_1, SyncMode.incremental),
+                _configured_stream(stream_2, SyncMode.incremental),
             ]
         )
 
         expected = [
             # stream 1 slice 1
             *_as_records("s1", stream_output),
-            _state({"s1": state}),
+            _as_state({"s1": state}, "s1", state) if per_stream_enabled else _as_state({"s1": state}),
             # stream 1 slice 2
             *_as_records("s1", stream_output),
-            _state({"s1": state}),
+            _as_state({"s1": state}, "s1", state) if per_stream_enabled else _as_state({"s1": state}),
             # stream 2 slice 1
             *_as_records("s2", stream_output),
-            _state({"s1": state, "s2": state}),
+            _as_state({"s1": state, "s2": state}, "s2", state) if per_stream_enabled else _as_state({"s1": state, "s2": state}),
             # stream 2 slice 2
             *_as_records("s2", stream_output),
-            _state({"s1": state, "s2": state}),
+            _as_state({"s1": state, "s2": state}, "s2", state) if per_stream_enabled else _as_state({"s1": state, "s2": state}),
         ]
 
-        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state=defaultdict(dict))))
+        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state=input_state)))
 
         assert expected == messages
 
-    def test_no_slices(self, mocker):
+    @pytest.mark.parametrize(
+        "use_legacy",
+        [
+            pytest.param(True, id="test_incoming_stream_state_as_legacy_format"),
+            pytest.param(False, id="test_incoming_stream_state_as_per_stream_format"),
+        ],
+    )
+    @pytest.mark.parametrize(
+        "per_stream_enabled",
+        [
+            pytest.param(True, id="test_source_emits_state_as_per_stream_format"),
+            pytest.param(False, id="test_source_emits_state_as_per_stream_format"),
+        ],
+    )
+    def test_no_slices(self, mocker, use_legacy, per_stream_enabled):
         """
         Tests that an incremental read returns at least one state messages even if no records were read:
             1. outputs a state message after reading the entire stream
         """
+        if use_legacy:
+            input_state = defaultdict(dict)
+        else:
+            input_state = []
         slices = []
         stream_output = [{"k1": "v1"}, {"k2": "v2"}, {"k3": "v3"}]
         state = {"cursor": "value"}
-        s1 = MockStreamWithState(
+        stream_1 = MockStreamWithState(
             [
                 (
                     {
@@ -479,7 +633,7 @@ class TestIncrementalRead:
             name="s1",
             state=state,
         )
-        s2 = MockStreamWithState(
+        stream_2 = MockStreamWithState(
             [
                 (
                     {
@@ -505,35 +659,51 @@ class TestIncrementalRead:
             return_value=2,
         )
 
-        src = MockSource(streams=[s1, s2])
+        src = MockSource(streams=[stream_1, stream_2], per_stream=per_stream_enabled)
         catalog = ConfiguredAirbyteCatalog(
             streams=[
-                _configured_stream(s1, SyncMode.incremental),
-                _configured_stream(s2, SyncMode.incremental),
+                _configured_stream(stream_1, SyncMode.incremental),
+                _configured_stream(stream_2, SyncMode.incremental),
             ]
         )
 
         expected = [
-            _state({"s1": state}),
-            _state({"s1": state, "s2": state}),
+            _as_state({"s1": state}, "s1", state) if per_stream_enabled else _as_state({"s1": state}),
+            _as_state({"s1": state, "s2": state}, "s2", state) if per_stream_enabled else _as_state({"s1": state, "s2": state}),
         ]
 
-        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state=defaultdict(dict))))
+        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state=input_state)))
 
-        print(f"expected:\n{expected}")
-        print(f"messages:\n{messages}")
         assert expected == messages
 
-    def test_with_slices_and_interval(self, mocker):
+    @pytest.mark.parametrize(
+        "use_legacy",
+        [
+            pytest.param(True, id="test_incoming_stream_state_as_legacy_format"),
+            pytest.param(False, id="test_incoming_stream_state_as_per_stream_format"),
+        ],
+    )
+    @pytest.mark.parametrize(
+        "per_stream_enabled",
+        [
+            pytest.param(True, id="test_source_emits_state_as_per_stream_format"),
+            pytest.param(False, id="test_source_emits_state_as_per_stream_format"),
+        ],
+    )
+    def test_with_slices_and_interval(self, mocker, use_legacy, per_stream_enabled):
         """
         Tests that an incremental read which uses slices and a checkpoint interval:
             1. outputs all records
             2. outputs a state message every N records (N=checkpoint_interval)
             3. outputs a state message after reading the entire slice
         """
+        if use_legacy:
+            input_state = defaultdict(dict)
+        else:
+            input_state = []
         slices = [{"1": "1"}, {"2": "2"}]
         stream_output = [{"k1": "v1"}, {"k2": "v2"}, {"k3": "v3"}]
-        s1 = MockStream(
+        stream_1 = MockStream(
             [
                 (
                     {
@@ -547,7 +717,7 @@ class TestIncrementalRead:
             ],
             name="s1",
         )
-        s2 = MockStream(
+        stream_2 = MockStream(
             [
                 (
                     {
@@ -573,11 +743,11 @@ class TestIncrementalRead:
             return_value=2,
         )
 
-        src = MockSource(streams=[s1, s2])
+        src = MockSource(streams=[stream_1, stream_2], per_stream=per_stream_enabled)
         catalog = ConfiguredAirbyteCatalog(
             streams=[
-                _configured_stream(s1, SyncMode.incremental),
-                _configured_stream(s2, SyncMode.incremental),
+                _configured_stream(stream_1, SyncMode.incremental),
+                _configured_stream(stream_2, SyncMode.incremental),
             ]
         )
 
@@ -585,29 +755,47 @@ class TestIncrementalRead:
             # stream 1 slice 1
             _as_record("s1", stream_output[0]),
             _as_record("s1", stream_output[1]),
-            _state({"s1": state}),
+            _as_state({"s1": state}, "s1", state) if per_stream_enabled else _as_state({"s1": state}),
             _as_record("s1", stream_output[2]),
-            _state({"s1": state}),
+            _as_state({"s1": state}, "s1", state) if per_stream_enabled else _as_state({"s1": state}),
             # stream 1 slice 2
             _as_record("s1", stream_output[0]),
             _as_record("s1", stream_output[1]),
-            _state({"s1": state}),
+            _as_state({"s1": state}, "s1", state) if per_stream_enabled else _as_state({"s1": state}),
             _as_record("s1", stream_output[2]),
-            _state({"s1": state}),
+            _as_state({"s1": state}, "s1", state) if per_stream_enabled else _as_state({"s1": state}),
             # stream 2 slice 1
             _as_record("s2", stream_output[0]),
             _as_record("s2", stream_output[1]),
-            _state({"s1": state, "s2": state}),
+            _as_state({"s1": state, "s2": state}, "s2", state) if per_stream_enabled else _as_state({"s1": state, "s2": state}),
             _as_record("s2", stream_output[2]),
-            _state({"s1": state, "s2": state}),
+            _as_state({"s1": state, "s2": state}, "s2", state) if per_stream_enabled else _as_state({"s1": state, "s2": state}),
             # stream 2 slice 2
             _as_record("s2", stream_output[0]),
             _as_record("s2", stream_output[1]),
-            _state({"s1": state, "s2": state}),
+            _as_state({"s1": state, "s2": state}, "s2", state) if per_stream_enabled else _as_state({"s1": state, "s2": state}),
             _as_record("s2", stream_output[2]),
-            _state({"s1": state, "s2": state}),
+            _as_state({"s1": state, "s2": state}, "s2", state) if per_stream_enabled else _as_state({"s1": state, "s2": state}),
         ]
 
-        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state=defaultdict(dict))))
+        messages = _fix_emitted_at(list(src.read(logger, {}, catalog, state=input_state)))
 
         assert expected == messages
+
+
+def test_checkpoint_state_from_stream_instance():
+    teams_stream = MockStreamOverridesStateMethod()
+    managers_stream = StreamNoStateMethod()
+    src = MockSource(streams=[teams_stream, managers_stream])
+    state_manager = ConnectorStateManager({"teams": teams_stream, "managers": managers_stream}, [])
+
+    # The stream_state passed to checkpoint_state() should be ignored since stream implements state function
+    teams_stream.state = {"updated_at": "2022-09-11"}
+    actual_message = src._checkpoint_state(teams_stream, {"ignored": "state"}, state_manager)
+    assert actual_message == _as_state({"teams": {"updated_at": "2022-09-11"}}, "teams", {"updated_at": "2022-09-11"})
+
+    # The stream_state passed to checkpoint_state() should be used since the stream does not implement state function
+    actual_message = src._checkpoint_state(managers_stream, {"updated": "expected_here"}, state_manager)
+    assert actual_message == _as_state(
+        {"teams": {"updated_at": "2022-09-11"}, "managers": {"updated": "expected_here"}}, "managers", {"updated": "expected_here"}
+    )
